import 'package:flutter_query/src/query_cache.dart';
import 'package:flutter_query/src/types.dart';

import '../flutter_query.dart';

typedef void QueryResultListener<T>(QueryResult<T> state);

class QueryObserver<T> {
  QueryConfig config;
  QueryCache cache;
  int staleTime = 0;

  QueryResultListener<T> _listener;
  Query<T> _query;
  // Future _staleTimer;

  QueryObserver(
    this.config,
    this.cache, {
    this.staleTime,
  }) {
    _query = cache.getQueryByKey(config.queryKey);

    if (_query == null) {
      _query = cache.createQuery(config);
    }
  }

  Function subscribe(QueryResultListener<T> listener) {
    _listener = listener;
    // if (_staleTimer != null) {
    //   _staleTimer =
    //       Future.delayed(Duration(milliseconds: staleTime)).then((value) {
    //     query.fetch();
    //   });
    // }
    final subscription = _query.stateStream.listen(onQueryUpdate);
    _query.fetch();

    return () {
      subscription.cancel();
      _listener = null;
    };
  }

  void onQueryUpdate(QueryState<T> state) {
    if (state.isSuccess && config.onSuccess != null) {
      config.onSuccess(state.data);
    } else if (state.isError && config.onError != null) {
      config.onError(state.data);
    }

    if ((state.isSuccess || state.isError) && config.onSettled != null) {
      config.onSettled(state.data);
    }

    _listener(QueryResult<T>(
      data: state.data,
      err: state.snapshot?.hasError ?? false ? state.snapshot?.error : null,
      failureCount: state.failureCount,
      isError: state.isError,
      isFetching: state.isFetching,
      isIdle: state.isIdle,
      isLoading: state.isLoading,
      isSuccess: state.isSuccess,
      isStale: false,
      fetchMore: _query.config.isInfinite ? () => _query.fetch() : () {},
      canFetchMore: state.canFetchMore,
      status: state.status,
    ));
  }
}

class QueryResult<T> {
  T data;
  dynamic err;
  int failureCount;
  bool isError;
  bool isFetchedAfterInit;
  bool isFetching;
  bool isIdle;
  bool isLoading;
  bool isPreviousData;
  bool isStale;
  bool isSuccess;
  bool canFetchMore;
  Function fetchMore;
  Function refetch;
  Function remove;
  QueryStatus status;

  QueryResult({
    this.data,
    this.err,
    this.failureCount,
    this.isError,
    this.isFetchedAfterInit,
    this.isFetching,
    this.isIdle,
    this.isLoading,
    this.isPreviousData,
    this.isStale,
    this.isSuccess,
    this.refetch,
    this.remove,
    this.status,
    this.canFetchMore,
    this.fetchMore,
  });
}
